// Copyright (c) 2023 PAL Robotics S.L. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// Copyright 2016, 2017, 2018, 2019, 2020, 2021 Erwan BERNARD.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// Copied and adapted from https://github.com/edmBernard/pybind11_opencv_numpy
// borrowed in spirit from https://github.com/yati-sagade/opencv-ndarray-conversion

#include "pyhri/ndarray_converter.h"

#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION
#include <numpy/ndarrayobject.h>

#if PY_VERSION_HEX >= 0x03000000
    #define PyInt_Check PyLong_Check
    #define PyInt_AsLong PyLong_AsLong
#endif

struct Tmp
{
  const char * name;

  Tmp(const char * _name)
  : name(_name) {}
};

Tmp info("return value");

bool NDArrayConverter::init_numpy()
{
  // this has to be in this file, since PyArray_API is defined as static
  import_array1(false);
  return true;
}

// The following conversion functions are taken/adapted from OpenCV's cv2.cpp file
// inside modules/python/src2 folder (OpenCV 4.5.2)


static PyObject * opencv_error = 0;

static int failmsg(const char * fmt, ...)
{
  char str[1000];

  va_list ap;
  va_start(ap, fmt);
  vsnprintf(str, sizeof(str), fmt, ap);
  va_end(ap);

  PyErr_SetString(PyExc_TypeError, str);
  return 0;
}

class PyAllowThreads
{
public:
  PyAllowThreads()
  : _state(PyEval_SaveThread()) {}
  ~PyAllowThreads()
  {
    PyEval_RestoreThread(_state);
  }

private:
  PyThreadState * _state;
};

class PyEnsureGIL
{
public:
  PyEnsureGIL()
  : _state(PyGILState_Ensure()) {}
  ~PyEnsureGIL()
  {
    PyGILState_Release(_state);
  }

private:
  PyGILState_STATE _state;
};

#define ERRWRAP2(expr) \
  try \
  { \
    PyAllowThreads allowThreads; \
    expr; \
  } \
  catch (const cv::Exception & e) \
  { \
    PyErr_SetString(opencv_error, e.what()); \
    return 0; \
  }

using namespace cv;

class NumpyAllocator : public MatAllocator
{
public:
  NumpyAllocator() {stdAllocator = Mat::getStdAllocator();}
  ~NumpyAllocator() {}

  UMatData * allocate(PyObject * o, int dims, const int * sizes, int type, size_t * step) const
  {
    UMatData * u = new UMatData(this);
    u->data = u->origdata = (uchar *)PyArray_DATA((PyArrayObject *) o);
    npy_intp * _strides = PyArray_STRIDES((PyArrayObject *) o);
    for (int i = 0; i < dims - 1; i++) {
      step[i] = (size_t)_strides[i];
    }
    step[dims - 1] = CV_ELEM_SIZE(type);
    u->size = sizes[0] * step[0];
    u->userdata = o;
    return u;
  }

#if CV_MAJOR_VERSION < 4
  UMatData * allocate(
    int dims0, const int * sizes, int type, void * data, size_t * step, int flags,
    UMatUsageFlags usageFlags) const
#else
  UMatData * allocate(
    int dims0, const int * sizes, int type, void * data, size_t * step,
    AccessFlag flags, UMatUsageFlags usageFlags) const
#endif
  {
    if (data != 0) {
      CV_Error(Error::StsAssert, "The data should normally be NULL!");
      // probably this is safe to do in such extreme case
      return stdAllocator->allocate(dims0, sizes, type, data, step, flags, usageFlags);
    }
    PyEnsureGIL gil;

    int depth = CV_MAT_DEPTH(type);
    int cn = CV_MAT_CN(type);
    const int f = (int)(sizeof(size_t) / 8);
    int typenum = depth == CV_8U ? NPY_UBYTE : depth == CV_8S ? NPY_BYTE :
      depth == CV_16U ? NPY_USHORT : depth == CV_16S ? NPY_SHORT :
      depth == CV_32S ? NPY_INT : depth == CV_32F ? NPY_FLOAT :
      depth == CV_64F ? NPY_DOUBLE : f * NPY_ULONGLONG + (f ^ 1) * NPY_UINT;
    int i, dims = dims0;
    cv::AutoBuffer<npy_intp> _sizes(dims + 1);
    for (i = 0; i < dims; i++) {
      _sizes[i] = sizes[i];
    }
    if (cn > 1) {
      _sizes[dims++] = cn;
    }
#if CV_MAJOR_VERSION >= 4 || (CV_MAJOR_VERSION == 3 && CV_VERSION_MINOR >= 5) || \
    (CV_MAJOR_VERSION == 3 && CV_VERSION_MINOR == 4 && CV_VERSION_REVISION >= 3)
    // Use cv::AutoBuffer::data() in OpenCV 3.4.3 and above
    PyObject * o = PyArray_SimpleNew(dims, _sizes.data(), typenum);
#else
    // Use older cv::AutoBuffer::operator _Tp*()
    PyObject * o = PyArray_SimpleNew(dims, &(_sizes[0]), typenum);
#endif
    if (!o) {
      CV_Error_(
        Error::StsError,
        ("The numpy array of typenum=%d, ndims=%d can not be created", typenum, dims));
    }
    return allocate(o, dims0, sizes, type, step);
  }

#if CV_MAJOR_VERSION < 4
  bool allocate(UMatData * u, int accessFlags, UMatUsageFlags usageFlags) const
#else
  bool allocate(UMatData * u, AccessFlag accessFlags, UMatUsageFlags usageFlags) const
#endif
  {
    return stdAllocator->allocate(u, accessFlags, usageFlags);
  }

  void deallocate(UMatData * u) const
  {
    if (!u) {
      return;
    }
    PyEnsureGIL gil;
    CV_Assert(u->urefcount >= 0);
    CV_Assert(u->refcount >= 0);
    if (u->refcount == 0) {
      PyObject * o = (PyObject *)u->userdata;
      Py_XDECREF(o);
      delete u;
    }
  }

  const MatAllocator * stdAllocator;
};

NumpyAllocator g_numpyAllocator;

bool NDArrayConverter::toMat(PyObject * o, Mat & m)
{
  bool allowND = true;
  if (!o || o == Py_None) {
    if (!m.data) {
      m.allocator = &g_numpyAllocator;
    }
    return true;
  }

  if (PyInt_Check(o) ) {
    double v[] = {static_cast<double>(PyInt_AsLong((PyObject *)o)), 0., 0., 0.};
    m = Mat(4, 1, CV_64F, v).clone();
    return true;
  }
  if (PyFloat_Check(o) ) {
    double v[] = {PyFloat_AsDouble((PyObject *)o), 0., 0., 0.};
    m = Mat(4, 1, CV_64F, v).clone();
    return true;
  }
  if (PyTuple_Check(o) ) {
    int i, sz = (int)PyTuple_Size((PyObject *)o);
    m = Mat(sz, 1, CV_64F);
    for (i = 0; i < sz; i++) {
      PyObject * oi = PyTuple_GetItem(o, i);
      if (PyInt_Check(oi) ) {
        m.at<double>(i) = (double)PyInt_AsLong(oi);
      } else if (PyFloat_Check(oi) ) {
        m.at<double>(i) = (double)PyFloat_AsDouble(oi);
      } else {
        failmsg("%s is not a numerical tuple", info.name);
        m.release();
        return false;
      }
    }
    return true;
  }

  if (!PyArray_Check(o) ) {
    failmsg("%s is not a numpy array, neither a scalar", info.name);
    return false;
  }

  PyArrayObject * oarr = (PyArrayObject *) o;

  bool needcopy = false, needcast = false;
  int typenum = PyArray_TYPE(oarr), new_typenum = typenum;
  int type = typenum == NPY_UBYTE ? CV_8U :
    typenum == NPY_BYTE ? CV_8S :
    typenum == NPY_USHORT ? CV_16U :
    typenum == NPY_SHORT ? CV_16S :
    typenum == NPY_INT ? CV_32S :
    typenum == NPY_INT32 ? CV_32S :
    typenum == NPY_FLOAT ? CV_32F :
    typenum == NPY_DOUBLE ? CV_64F : -1;

  if (type < 0) {
    if (typenum == NPY_INT64 || typenum == NPY_UINT64 || typenum == NPY_LONG) {
      needcopy = needcast = true;
      new_typenum = NPY_INT;
      type = CV_32S;
    } else {
      failmsg("%s data type = %d is not supported", info.name, typenum);
      return false;
    }
  }

#ifndef CV_MAX_DIM
  const int CV_MAX_DIM = 32;
#endif

  int ndims = PyArray_NDIM(oarr);
  if (ndims >= CV_MAX_DIM) {
    failmsg("%s dimensionality (=%d) is too high", info.name, ndims);
    return false;
  }

  int size[CV_MAX_DIM + 1];
  size_t step[CV_MAX_DIM + 1];
  size_t elemsize = CV_ELEM_SIZE1(type);
  const npy_intp * _sizes = PyArray_DIMS(oarr);
  const npy_intp * _strides = PyArray_STRIDES(oarr);
  bool ismultichannel = ndims == 3 && _sizes[2] <= CV_CN_MAX;

  for (int i = ndims - 1; i >= 0 && !needcopy; i--) {
    // these checks handle cases of
    //  a) multi-dimensional (ndims > 2) arrays, as well as simpler 1- and 2-dimensional cases
    //  b) transposed arrays, where _strides[] elements go in non-descending order
    //  c) flipped arrays, where some of _strides[] elements are negative
    // the _sizes[i] > 1 is needed to avoid spurious copies when NPY_RELAXED_STRIDES is set
    if ( (i == ndims - 1 && _sizes[i] > 1 && (size_t)_strides[i] != elemsize) ||
      (i < ndims - 1 && _sizes[i] > 1 && _strides[i] < _strides[i + 1]) )
    {
      needcopy = true;
    }
  }

  if (ismultichannel && _strides[1] != (npy_intp)elemsize * _sizes[2]) {
    needcopy = true;
  }

  if (needcopy) {
    // if (info.outputarg)
    // {
    //     failmsg("Layout of the output array %s is incompatible with cv::Mat (step[ndims-1] != elemsize or step[1] != elemsize*nchannels)", info.name);
    //     return false;
    // }

    if (needcast) {
      o = PyArray_Cast(oarr, new_typenum);
      oarr = (PyArrayObject *) o;
    } else {
      oarr = PyArray_GETCONTIGUOUS(oarr);
      o = (PyObject *) oarr;
    }

    _strides = PyArray_STRIDES(oarr);
  }

  // Normalize strides in case NPY_RELAXED_STRIDES is set
  size_t default_step = elemsize;
  for (int i = ndims - 1; i >= 0; --i) {
    size[i] = (int)_sizes[i];
    if (size[i] > 1) {
      step[i] = (size_t)_strides[i];
      default_step = step[i] * size[i];
    } else {
      step[i] = default_step;
      default_step *= size[i];
    }
  }

  // handle degenerate case
  if (ndims == 0) {
    size[ndims] = 1;
    step[ndims] = elemsize;
    ndims++;
  }

  if (ismultichannel) {
    ndims--;
    type |= CV_MAKETYPE(0, size[2]);
  }

  if (ndims > 2 && !allowND) {
    failmsg("%s has more than 2 dimensions", info.name);
    return false;
  }

  m = Mat(ndims, size, type, PyArray_DATA(oarr), step);
  m.u = g_numpyAllocator.allocate(o, ndims, size, type, step);
  m.addref();

  if (!needcopy) {
    Py_INCREF(o);
  }
  m.allocator = &g_numpyAllocator;

  return true;
}

PyObject * NDArrayConverter::toNDArray(const cv::Mat & m)
{
  if (!m.data) {
    Py_RETURN_NONE;
  }
  Mat temp, * p = (Mat *)&m;
  if (!p->u || p->allocator != &g_numpyAllocator) {
    temp.allocator = &g_numpyAllocator;
    ERRWRAP2(m.copyTo(temp));
    p = &temp;
  }
  PyObject * o = (PyObject *)p->u->userdata;
  Py_INCREF(o);
  return o;
}
